/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use super::{Assign, From, Number};

use num_traits::{Num, One, Zero};
use rug::{self, ops::Pow, ops::PowAssign, rational::ParseRationalError};
use rug::Assign as RugAssign;

use std::cmp::{Ord, Ordering, PartialEq, PartialOrd};
use std::fmt;
use std::ops;

#[derive(Clone, PartialEq, PartialOrd)]
pub struct Rational(rug::Rational);

impl Number for Rational {
	fn new() -> Self { Self(rug::Rational::new()) }
	
	fn describe() -> String { "--numbers rational".to_string() }
	fn describe_opt() -> String { String::new() }
	
	fn pow_assign(&mut self, exponent: i32) {
		self.0.pow_assign(exponent);
	}
	
	fn floor_mut(&mut self, dps: usize) {
		if dps == 0 {
			self.0.floor_mut();
		} else {
			let factor = rug::Rational::from(10).pow(dps as u32);
			self.0 *= &factor;
			self.0.floor_mut();
			self.0 /= factor;
		}
	}
	
	fn ceil_mut(&mut self, dps: usize) {
		if dps == 0 {
			self.0.ceil_mut();
		} else {
			let factor = rug::Rational::from(10).pow(dps as u32);
			self.0 *= &factor;
			self.0.ceil_mut();
			self.0 /= factor;
		}
	}
}

impl Num for Rational {
	type FromStrRadixErr = ParseRationalError;
	fn from_str_radix(str: &str, radix: u32) -> Result<Self, Self::FromStrRadixErr> {
		match rug::Rational::parse_radix(str, radix as i32) {
			Ok(value) => Ok(Self(rug::Rational::from(value))),
			Err(err) => Err(err)
		}
	}
}

impl Assign for Rational {
	fn assign(&mut self, src: Self) { self.0.assign(src.0) }
}

impl Assign<&Self> for Rational {
	fn assign(&mut self, src: &Self) { self.0.assign(&src.0) }
}

impl From<usize> for Rational {
	fn from(n: usize) -> Self { Self(rug::Rational::from(n)) }
}

impl From<f64> for Rational {
	fn from(n: f64) -> Self {
		// FIXME: This is very broken!
		return Self((rug::Rational::from_f64(n).unwrap() * rug::Rational::from(100000)).round() / rug::Rational::from(100000));
	}
}

impl fmt::Display for Rational {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		if let Some(precision) = f.precision() {
			if precision == 0 {
				let result = rug::Integer::from((&self.0).round_ref()).to_string();
				return f.write_str(&result);
			} else {
				let base = rug::Rational::from(10).pow(precision as u32);
				let mut result = rug::Integer::from((&self.0 * base).abs().round_ref()).to_string();
				
				let should_add_minus = (self.0 < 0) && result != "0";
				
				// Add leading 0s
				result = format!("{0:0>1$}", result, precision + 1);
				
				// Add the decimal point
				result.insert(result.len() - precision, '.');
				
				// Add the sign
				if should_add_minus {
					result.insert(0, '-');
				}
				
				return f.write_str(&result);
			}
		} else {
			return self.0.fmt(f);
		}
	}
}

impl One for Rational {
	fn one() -> Self { Self(rug::Rational::from(1)) }
}

impl Zero for Rational {
	fn zero() -> Self { Self::new() }
	fn is_zero(&self) -> bool { self.0 == rug::Rational::new() }
}

impl Eq for Rational {}
impl Ord for Rational {
	fn cmp(&self, other: &Self) -> Ordering { self.0.cmp(&other.0) }
}

impl ops::Neg for Rational {
	type Output = Self;
	fn neg(self) -> Self::Output { Self(-self.0) }
}

impl ops::Add for Rational {
	type Output = Self;
	fn add(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Sub for Rational {
	type Output = Self;
	fn sub(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Mul for Rational {
	type Output = Self;
	fn mul(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Div for Rational {
	type Output = Self;
	fn div(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Rem for Rational {
	type Output = Self;
	fn rem(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Add<&Self> for Rational {
	type Output = Self;
	fn add(self, rhs: &Self) -> Self::Output { Self(self.0 + &rhs.0) }
}

impl ops::Sub<&Self> for Rational {
	type Output = Self;
	fn sub(self, rhs: &Self) -> Self::Output { Self(self.0 - &rhs.0) }
}

impl ops::Mul<&Self> for Rational {
	type Output = Self;
	fn mul(self, rhs: &Self) -> Self::Output { Self(self.0 * &rhs.0) }
}

impl ops::Div<&Self> for Rational {
	type Output = Self;
	fn div(self, rhs: &Self) -> Self::Output { Self(self.0 / &rhs.0) }
}

impl ops::Rem<&Self> for Rational {
	type Output = Self;
	fn rem(self, _rhs: &Self) -> Self::Output {
		todo!()
	}
}

impl ops::AddAssign for Rational {
	fn add_assign(&mut self, rhs: Self) { self.0 += rhs.0; }
}

impl ops::SubAssign for Rational {
	fn sub_assign(&mut self, rhs: Self) { self.0 -= rhs.0; }
}

impl ops::MulAssign for Rational {
	fn mul_assign(&mut self, rhs: Self) { self.0 *= rhs.0; }
}

impl ops::DivAssign for Rational {
	fn div_assign(&mut self, rhs: Self) { self.0 /= rhs.0; }
}

impl ops::RemAssign for Rational {
	fn rem_assign(&mut self, _rhs: Self) {
		todo!()
	}
}

impl ops::AddAssign<&Self> for Rational {
	fn add_assign(&mut self, rhs: &Self) { self.0 += &rhs.0; }
}

impl ops::SubAssign<&Self> for Rational {
	fn sub_assign(&mut self, rhs: &Self) { self.0 -= &rhs.0; }
}

impl ops::MulAssign<&Self> for Rational {
	fn mul_assign(&mut self, rhs: &Self) { self.0 *= &rhs.0; }
}

impl ops::DivAssign<&Self> for Rational {
	fn div_assign(&mut self, _rhs: &Self) {
		todo!()
	}
}

impl ops::RemAssign<&Self> for Rational {
	fn rem_assign(&mut self, _rhs: &Self) {
		todo!()
	}
}

impl ops::Neg for &Rational {
	type Output = Rational;
	fn neg(self) -> Self::Output { Rational(rug::Rational::from(-&self.0)) }
}

impl ops::Add<Self> for &Rational {
	type Output = Rational;
	fn add(self, rhs: Self) -> Self::Output { Rational(rug::Rational::from(&self.0 + &rhs.0)) }
}

impl ops::Sub<Self> for &Rational {
	type Output = Rational;
	fn sub(self, rhs: Self) -> Self::Output { Rational(rug::Rational::from(&self.0 - &rhs.0)) }
}

impl ops::Mul<Self> for &Rational {
	type Output = Rational;
	fn mul(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Div<Self> for &Rational {
	type Output = Rational;
	fn div(self, rhs: Self) -> Self::Output { Rational(rug::Rational::from(&self.0 / &rhs.0)) }
}

impl ops::Rem<Self> for &Rational {
	type Output = Rational;
	fn rem(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

/*
impl ops::Add<&&Rational> for &Rational {
	
}

impl ops::Sub<&&Rational> for &Rational {
	
}

impl ops::Mul<&&Rational> for &Rational {
	
}

impl ops::Div<&&Rational> for &Rational {
	
}

impl ops::Rem<&&Rational> for &Rational {
	
}
*/
