/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

mod utils;

use opentally::election::{CandidateState, CountState, Election};
use opentally::numbers::Rational;
use opentally::stv;

use csv::StringRecord;

use std::fs::File;
use std::io::{self, BufRead};

#[test]
fn ers97_rational() {
	let stv_opts = stv::STVOptions {
		round_tvs: Some(2),
		round_weights: Some(2),
		round_votes: Some(2),
		round_quota: Some(2),
		sum_surplus_transfers: stv::SumSurplusTransfersMode::SingleStep,
		normalise_ballots: false,
		quota: stv::QuotaType::DroopExact,
		quota_criterion: stv::QuotaCriterion::GreaterOrEqual,
		quota_mode: stv::QuotaMode::ERS97,
		ties: vec![],
		surplus: stv::SurplusMethod::EG,
		surplus_order: stv::SurplusOrder::BySize,
		transferable_only: true,
		exclusion: stv::ExclusionMethod::ByValue,
		bulk_exclude: true,
		defer_surpluses: true,
		pp_decimals: 2,
	};
	
	// ---------------------------------------------
	// Custom implementation due to nontransferables
	// and vote required for election
	
	// Read CSV file
	let reader = csv::ReaderBuilder::new()
		.has_headers(false)
		.from_path("tests/data/ers97.csv")
		.expect("IO Error");
	let records: Vec<StringRecord> = reader.into_records().map(|r| r.expect("Syntax Error")).collect();
	
	let mut candidates: Vec<&str> = records.iter().skip(2).map(|r| &r[0]).collect();
	// Remove NT/VRE rows
	candidates.truncate(candidates.len() - 2);
	
	// TODO: Validate candidate names
	
	let stages: Vec<usize> = records.first().unwrap().iter().skip(1).step_by(2).map(|s| s.parse().unwrap()).collect();
	
	// Read BLT
	let file = File::open("tests/data/ers97.blt").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	
	let election: Election<Rational> = Election::from_blt(lines.map(|r| r.expect("IO Error").to_string()).into_iter());
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Distribute first preferences
	stv::count_init(&mut state, &stv_opts);
	let mut stage_num = 1;
	
	for (idx, stage) in stages.into_iter().enumerate() {
		while stage_num < stage {
			// Step through stages
			// Assert count not yet done
			assert_eq!(stv::count_one_stage(&mut state, &stv_opts).unwrap(), false);
			stage_num += 1;
		}
		
		println!("Col at idx {}", idx);
		
		let mut candidate_votes: Vec<Option<Rational>> = records.iter().skip(2)
			.map(|r|
				if r[idx*2 + 1].len() > 0 {
					Some(opentally::numbers::From::from(r[idx*2 + 1].parse::<f64>().expect("Syntax Error")))
				} else {
					None
				})
			.collect();
		
		// Validate NT/VRE
		let vre_votes = candidate_votes.pop().unwrap();
		let nt_votes = candidate_votes.pop().unwrap();
		
		assert!(&state.exhausted.votes + &state.loss_fraction.votes == nt_votes.unwrap());
		if let Some(v) = vre_votes {
			assert!(state.vote_required_election.as_ref().unwrap() == &v);
		}
		
		// Remove NT/VRE rows
		candidate_votes.truncate(candidate_votes.len() - 2);
		
		// Validate candidate votes
		for (candidate, votes) in state.election.candidates.iter().zip(candidate_votes) {
			let count_card = state.candidates.get(candidate).unwrap();
			assert!(&count_card.votes == votes.as_ref().unwrap(), "Failed to validate votes for candidate {}. Expected {:}, got {:}", candidate.name, votes.unwrap(), count_card.votes);
		}
		
		// Validate candidate states
		let mut candidate_states: Vec<&str> = records.iter().skip(2).map(|r| &r[idx*2 + 2]).collect();
		candidate_states.truncate(candidate_states.len() - 2);
		
		for (candidate, candidate_state) in state.election.candidates.iter().zip(candidate_states) {
			let count_card = state.candidates.get(candidate).unwrap();
			if candidate_state == "" {
				assert!(count_card.state == CandidateState::Hopeful);
			} else if candidate_state == "EL" || candidate_state == "PEL" {
				assert!(count_card.state == CandidateState::Elected);
			} else if candidate_state == "EX" || candidate_state == "EXCLUDING" {
				assert!(count_card.state == CandidateState::Excluded);
			} else {
				panic!("Unknown state descriptor {}", candidate_state);
			}
		}
	}
}
